#!/usr/bin/python3
import csv
import math
import cmath
import time
import numpy
from argparse import ArgumentParser
import pandas as pd
import matplotlib.pyplot as plt
import climate_chamber
import VNA
import virtual_time
import json
import MeasurementPlot
import sys
import analysis
import almemo710
import PostPlot
import os
import pyvisa
import socket


TEMPERATURE_STABLE = 0x1
HUMIDITY_STABLE = 0x2
MAGNITUDE_STABLE = 0x4
PHASE_STABLE = 0x8
MEASUREMENT_STABLE = 0x10


# Class has only attributes which we are using in the read_data_function to read data from VNA and Chamber
class MeasurementData:
    
    # values for percentage temp_heater, percentage humidity heater, temperature sensor0, temperature sensor1, air pressure, Michael
    # values for measurement instruments temperature, measurement instruments humidity 
    def __init__(self, timestamp, temp, hum, power, frequency, s11, s21, s12, s22, perc_temp_heater, \
                 perc_hum_heater, temp_dut, temp_room, temp_meas_instr, hum_dut, hum_room, \
                 hum_meas_instr, air_press_room):
        self.timestamp = timestamp
        self.temp = temp
        self.hum = hum
        self.power = power
        self.frequency = frequency
        self.s11 = s11
        self.s21 = s21
        self.s12 = s12
        self.s22 = s22
        # new variables for percentage temp heater and percentage humidity heater
        self.perc_temp_heater = perc_temp_heater
        self.perc_hum_heater = perc_hum_heater
        # new variables for DUT temp, room temp, Michael
        # new variable for measurement instruments temp, Michael
        self.temp_dut = temp_dut
        self.temp_room = temp_room
        self.temp_meas_instr = temp_meas_instr
        # new variables for DUT humidity, room humidity, air pressure room, Michael
        self.hum_dut = hum_dut
        self.hum_room = hum_room
        self.air_press_room = air_press_room
        # new variable for measurement instruments humidity, Michael 
        self.hum_meas_instr = hum_meas_instr


class Measurements:
    def __init__(self, chamber_address, vna_address, output_basename, standby, config_data, \
                 ext_sensor_channels, logger_adress):
        self.max_delta_temp = config_data['delta_temp']
        self.max_delta_hum = config_data['delta_hum']
        self.max_delta_mag = config_data['delta_mag']
        self.max_delta_phase = config_data['delta_phase']
        self.ext_sensor_channels = ext_sensor_channels
        self.sleep_time = config_data["sleep_time"]
        self.frequency = config_data["frequency"]
        self.vna_config_file = config_data["vna_config_file"]
        target_accuracy = [self.max_delta_temp, self.max_delta_hum]
        self.chamber = climate_chamber.create_chamber(chamber_address, target_accuracy)
        self.vna = VNA.create_vna(vna_address, target_accuracy)
        # new object for external sensors, decice 'ALMEMO710', fixed IP in code for device, Michael
        self.ext_sensors = almemo710.almemo710(ip = logger_adress, timeout = 1)
        self.standby = standby
        self.output_basename = output_basename
        self.clock = virtual_time.get_clock(chamber_address, target_accuracy)
        self.temperature_stable = False
        self.humidity_stable = False
        self.magnitude_stable = False
        self.phase_stable = False

        self.vna.load_config(self.vna_config_file, self.frequency)
        self.vna.create_new_trace("Trace1", "S11")
        self.vna.create_new_trace("Trace2", "S12")
        self.vna.create_new_trace("Trace3", "S21")
        self.vna.create_new_trace("Trace4", "S22")
        self.postplot_obj = None
        
        # request all sensor channels from ALMEMO710 of all sensors that are connected to this device, Michael
        self.ext_sensors.request_sens_channel_list()

        self.measurement_plot = MeasurementPlot.MeasurementPlot()
        self.data_collection = []

    def perform_measurements(self, sweep_file):
        with open(sweep_file) as file:
            try:
                measurement_number = 0
                while line := file.readline().rstrip():
                    list_of_values = line.split()
                    next_temp = float(list_of_values[0])  # target_temperature
                    next_hum = float(list_of_values[1])  # target_humidity
                    next_soaking = int(list_of_values[2])  # soaking_time
                    next_reads = int(list_of_values[3])  # number of stable readings

                    self.measurement_plot.fig.suptitle("Measurement "+str(measurement_number)+': ' + str(next_temp) +
                                                       ' degC, ' + str(next_hum) + ' rel. hum.', color="red")
                    self.perform_single_measurement(self.output_basename+'_'+str(measurement_number)+'.csv', next_temp, next_hum,
                                                    next_soaking, next_reads)
                    measurement_number += 1

            except KeyboardInterrupt:
                pass
            except MeasurementPlot.PlottingError:
                # Just the plotting failed, probably because the window was closed
                # The measurement was partly done and should be plottet, so the following code has to see the correct number of successful measurements
                measurement_number += 1

        plt.close()

        if self.standby:
            standby_response = self.chamber.set_mode('STANDBY')
            print(standby_response)

        return measurement_number

    def perform_sweep(self, start_val, stop_val, step_val, const_param, soaking_time, n_stable_reads, sweep_type):
        current_val = start_val
        sweep_values = []
        while current_val < stop_val:
            sweep_values.append(current_val)
            current_val = round(current_val + step_val, 1)
        sweep_values.append(stop_val)

        successful_sweep_values = []
        try:
            for val in sweep_values:
                if sweep_type == 'temperature':
                    target_temp = val
                    target_hum = const_param
                elif sweep_type == 'humidity':
                    target_temp = const_param
                    target_hum = val
                else:
                    raise Exception('Unknown sweep type: '+sweep_type)

                output_name = self.output_basename + '_'+str(target_temp)+'deg_'+str(target_hum)+'rh.csv'
                self.measurement_plot.fig.suptitle(sweep_type + " sweep measurement "+str(sweep_values.index(val)+1)+
                                                   '/'+str(len(sweep_values))+': ' + str(target_temp)+' degC, '
                                                   + str(target_hum) + ' rel. hum.',
                                                   color="red")
                self.perform_single_measurement(output_name, target_temp, target_hum, soaking_time, n_stable_reads)

        except KeyboardInterrupt:
            pass
        except MeasurementPlot.PlottingError:
            # Remove the remaining measurements from the list.
            # One measurement was partly done and should be plottet, so we leave 'val' in the list
            del sweep_values[sweep_values.index(val)+1 : ]



        plt.close()
        return sweep_values

    def perform_single_measurement(self, output, target_temp, target_hum, soaking_time, n_stable_reads):
        with open(output, mode='w', newline='') as csv_file:
            # fieldnames for percentage temperature heater and percntage humidity heater,
            # DUT Temperature, DUT humidity, room temperature, room humidity, airpressure room added,
            # measurement instruments temperature, measurement instruments humidity, Michael
            fieldnames = ['TIMESTAMP', 'TARGET_TEMPERATURE', 'READBACK_TEMPERATURE', 'TARGET_HUMIDITY',
                          'READBACK_HUMIDITY', 'RF_POWER', 'RF_FREQUENCY', 'DUT_IDENTIFIER', 'RUN_ID',
                          'EQUILIBRIUM_INDICATOR', 'TEMP_HEATER', 'HUM_HEATER',
                          'TEMP_DUT', 'TEMP_ROOM', 'TEMP_MEAS_INSTR',
                          'HUM_DUT','HUM_ROOM', 'HUM_MEAS_INSTR', 'AIR_PRESS_ROOM',
                          'S11_MAGNITUDE', 'S11_PHASE', 'S12_MAGNITUDE',
                          'S12_PHASE', 'S21_MAGNITUDE', 'S21_PHASE', 'S22_MAGNITUDE', 'S22_PHASE']
            # csv.dict writer add adda row wise
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
            writer.writeheader()
            self.data_collection = []
            plt.ion()

            set_const_response = self.chamber.set_const((target_temp, target_hum))
            print(set_const_response)
            set_mode_response = self.chamber.set_mode('CONSTANT')
            print(set_mode_response)

            number_of_soaking_reads = soaking_time / self.sleep_time + 1

            do_another_measurement = True
            #next_read_time is the starttime of the second read (i.e. read after this one). The time of
            #this read (i.e. the first read in the measurement) is now().
            next_read_time = self.clock.time() + self.sleep_time
            while do_another_measurement:
                # wait until set point is reached (+soaking time)
                magnitudes_queue = []
                phase_queue = []

                while True:
                    data = self.read_data()
                    # if it is not within the target range reset start time
                    self.temperature_stable = False
                    self.humidity_stable = False

                    self.temperature_stable = self.calculate_temperature_stability(target_temp, float(data.temp))
                    self.humidity_stable = self.calculate_humidity_stability(target_hum, float(data.hum))

                    # The queue must not be longer than the max number of soaking reads.
                    # If the queue is already full, we have to pop the first element before we can add the
                    # current measurement.
                    if len(magnitudes_queue) >= number_of_soaking_reads:
                        magnitudes_queue.pop(0)
                    if self.temperature_stable and self.humidity_stable:
                        magnitudes_queue.append(self.calculate_mean_magnitude(data.s21))
                    else:
                        magnitudes_queue.clear()
                    # check cable stability parameters
                    self.magnitude_stable = False
                    if len(magnitudes_queue) >= number_of_soaking_reads:
                        spread = max(magnitudes_queue) - min(magnitudes_queue)
                        if spread < 2*self.max_delta_mag:
                            self.magnitude_stable = True

                    if len(phase_queue) >= number_of_soaking_reads:
                        phase_queue.pop(0)
                    if self.temperature_stable and self.humidity_stable:
                        phase_queue.append(self.calculate_mean_phase(data.s21))
                    else:
                        phase_queue.clear()

                    self.phase_stable = False
                    if len(phase_queue) >= number_of_soaking_reads:
                        spread = max(phase_queue) - min(phase_queue)
                        if spread < 2*self.max_delta_phase:
                            self.phase_stable = True

                    print('Setpoint: ' + str(target_temp) + ' ' + str(target_hum) + ' | Temp: ' + data.temp +
                          ' °C' + ' | Humid: ' + data.hum + '%'
                          + ' | soaking read nr' + str(len(magnitudes_queue)))
                    self.store_and_plot_data(target_temp, target_hum, data, self.cook_up_equi_indicator())
                    writer.writerow(self.data_collection[-1])

                    if self.temperature_stable and self.humidity_stable and self.magnitude_stable and\
                            self.phase_stable:
                        reference_magnitude = magnitudes_queue[-1]
                        reference_phase = phase_queue[-1]
                        print('SOAKING FINISHED!')
                        break
                    else:
                        self.sleep_until(next_read_time)
                        next_read_time += self.sleep_time

                # perform the configured number of measurements and check that they are really stable
                # It started running after everything become stable
                supposedly_stable_measurements = []
                all_measurements_stable = True
                for i in range(0, n_stable_reads):
                    data = self.read_data()
                    self.temperature_stable = self.calculate_temperature_stability(target_temp, float(data.temp))
                    self.humidity_stable = self.calculate_humidity_stability(target_hum, float(data.hum))
                    mag = self.calculate_mean_magnitude(data.s21)
                    phase = self.calculate_mean_phase(data.s21)
                    self.magnitude_stable = (reference_magnitude-self.max_delta_mag <= mag) and\
                                            (mag <= reference_magnitude+self.max_delta_mag)
                    self.phase_stable = (reference_phase-self.max_delta_phase <= phase) and\
                                        (phase <= reference_phase+self.max_delta_phase)
                    self.store_and_plot_data(target_temp, target_hum, data, self.cook_up_equi_indicator())
                    supposedly_stable_measurements.append(self.data_collection[-1])
                                
                    if (self.temperature_stable and self.humidity_stable and self.magnitude_stable and
                            self.phase_stable):
                        print('Stable measurement ' + str(i+1) + '/' + str(n_stable_reads))
                        self.sleep_until(next_read_time)
                        next_read_time += self.sleep_time
                    else:
                        all_measurements_stable = False
                        print('Measurement not stable. Retrying.')
                        break

                for measurement in supposedly_stable_measurements:
                    if all_measurements_stable:
                        measurement['EQUILIBRIUM_INDICATOR'] = TEMPERATURE_STABLE | HUMIDITY_STABLE |\
                                                               MAGNITUDE_STABLE | PHASE_STABLE |\
                                                               MEASUREMENT_STABLE
                        do_another_measurement = False

                    writer.writerow(measurement)

    def sleep_until(self, wakeup_time):
        remaining_sleep_time = wakeup_time - self.clock.time()
        if remaining_sleep_time > 0:
            self.clock.sleep(remaining_sleep_time)

    def read_data(self):
        [temp, hum, mode, alarms] = self.chamber.read_monitor().split(',')
        # read percentage temperature heater, humidity heater from climate chamber, Michael
        perc_temp_heater, perc_hum_heater = self.chamber.get_heater_percentage()
        # loop with 10 Itherations in case of VNA readout error, Michael
        iteration = 0
        while iteration < 10:
            try:
                power = self.vna.get_current_power()
                frequency = self.vna.get_current_cw_frequency()
                self.vna.do_single_sweep()
                s11 = self.get_trace_data("Trace1")
                s12 = self.get_trace_data("Trace2")
                s21 = self.get_trace_data("Trace3")
                s22 = self.get_trace_data("Trace4")
                # if readout is successful, leave loop, Michael
                break
            # exception for pyvisa error or timeout
            except pyvisa.errors.VisaIOError:
                # call reset function for VNA statstus register and error queue
                self.vna.reset_status()
                print('An error occured during VNA read out')
                # wait one second for next try, Michael
                time.sleep(1)
                iteration+=1
            
        # loop with 3 Itherations in case of ALMEMO 710 readout error, Michael
        while iteration < 3:
            
            try:
                # request temperatures, humidities and and air pressure from external sensors, Michael
                temp_dut, temp_room, temp_meas_instr, hum_dut, hum_room, hum_meas_instr, \
                    air_press_room = self.get_ext_sensor_values()
                # if readout successfull, leave loop, Michael
                break
            # exception for telnet socket timeout
            except socket.timeout:
               
                print('An error occured during ALMEMO read out')
                # wait one second for next try, Michael
                time.sleep(1)
                iteration+=1
        
        
        # values percentage temp_heater, percentage hum_heater, temp_dut, temp_room, temp_meas_instr,
        # hum_dut, hum_room, hum_meas_instr air_press_room added to parameterlist in returned
        # values, Michael
        return MeasurementData(int(self.clock.time()), temp, hum, power, frequency, s11, s21, s12, s22, \
                               perc_temp_heater, perc_hum_heater, temp_dut, temp_room, temp_meas_instr, \
                                   hum_dut, hum_room, hum_meas_instr, air_press_room)

    def store_and_plot_data(self, target_temp, target_hum, data, equi_indicator):
        measurement = {
            'TIMESTAMP': data.timestamp,
            'TARGET_TEMPERATURE': target_temp,
            'READBACK_TEMPERATURE': float(data.temp),
            'TARGET_HUMIDITY': target_hum,
            'READBACK_HUMIDITY': float(data.hum),
            'RF_POWER': data.power,
            'RF_FREQUENCY': data.frequency,
            'EQUILIBRIUM_INDICATOR': equi_indicator,
            # percentage temperature heater added, Michael
            'TEMP_HEATER': data.perc_temp_heater,
            # percentage humidity heater added, Michael
            'HUM_HEATER': data.perc_hum_heater,
            # Temperature for DUT of external temp sensors added, Michael
            'TEMP_DUT': data.temp_dut,
            # Temperature for room of external temp sensors added, Michael
            'TEMP_ROOM': data.temp_room,
            # Temperature for measure instruments of external temp sensors added, Michael
            'TEMP_MEAS_INSTR': data.temp_meas_instr,      
            # Humidity for DUT of external hum sensors added, Michael
            'HUM_DUT': data.hum_dut,
            # Humidity for room of external hum sensors added, Michael         
            'HUM_ROOM': data.hum_room,
            # Humidity for measurement instruments of external hum sensors added, Michael         
            'HUM_MEAS_INSTR': data.hum_meas_instr,            
            # Air pressure for room of external sensors added, Michael
            'AIR_PRESS_ROOM': data.air_press_room,
            'S11_PHASE': self.calculate_mean_phase(data.s11),
            'S11_MAGNITUDE': 20*math.log10(self.calculate_mean_magnitude(data.s11)),
            'S21_PHASE': self.calculate_mean_phase(data.s21),
            'S21_MAGNITUDE': 20*math.log10(self.calculate_mean_magnitude(data.s21)),
            'S12_PHASE': self.calculate_mean_phase(data.s12),
            'S12_MAGNITUDE': 20*math.log10(self.calculate_mean_magnitude(data.s12)),
            'S22_PHASE': self.calculate_mean_phase(data.s22),
            'S22_MAGNITUDE': 20*math.log10(self.calculate_mean_magnitude(data.s22))
        }
        self.data_collection.append(measurement)
        data_frame = pd.DataFrame(self.data_collection)
        self.measurement_plot.draw(data_frame)

    def current_milli_time(self):
        return int(round(self.clock.time() * 1000))

    def get_trace_data(self, trace):
        return self.vna.get_list_of_measurement_values(trace, "SDAT")
    
    # new method to request temperaure values from external sensors of ahlborn, Michael 
    def get_ext_sensor_values(self):
        
        # empty dictonary for sensor values        
        sens_vals_dict = dict()
        
        # request temperatures, humidities and and air pressure from external sensors, Michael
        self.ext_sensors.request_meas_vals_all_channels()
        
        # request sensor value vor channel in dictonary and add item to dictionary of sensor values
        for key, value in self.ext_sensor_channels.items():
            
            sens_vals_dict[key] = self.ext_sensors.fetch_channel_param_from_meas_buffer(pattern = \
                                                    value)[0].meas_val
             
        # set values from series to variables
        sens_vals_series = pd.Series(sens_vals_dict)

        # set series items for sensor value to matching variable
        temp_dut =  sens_vals_series.temp_dut
        temp_room =  sens_vals_series.temp_room
        temp_meas_instr =  sens_vals_series.temp_meas_instr
        hum_dut =  sens_vals_series.hum_dut
        hum_room =  sens_vals_series.hum_room
        hum_meas_instr =  sens_vals_series.hum_meas_instr
        air_press_room = sens_vals_series.air_press_room
                 
        return temp_dut, temp_room, temp_meas_instr, hum_dut, hum_room, hum_meas_instr, \
            air_press_room
    

    def calculate_complex_numbers(self, values_list):
        real_num = values_list[::2]
        imaginary_num = values_list[1::2]
        complex_numbers = []
        for i, q in zip(real_num, imaginary_num):
            complex_numbers.append(complex(i, q))
        return complex_numbers

    def calculate_magnitudes(self, values_list):
        complex_numbers = self.calculate_complex_numbers(values_list)
        magnitudes = [abs(val) for val in complex_numbers]
        return magnitudes

    def calculate_mean_magnitude(self, values_list):
        magnitudes = self.calculate_magnitudes(values_list)
        return numpy.mean(magnitudes)

    def calculate_phases(self, values_list):
        complex_numbers = self.calculate_complex_numbers(values_list)
        phases = [math.degrees(cmath.phase(val)) for val in complex_numbers]
        return phases

    def calculate_mean_phase(self, values_list):
        phases = self.calculate_phases(values_list)
        return numpy.mean(phases)

    def cook_up_equi_indicator(self):
        equilibrium_indicator = 0

        if self.temperature_stable:
            equilibrium_indicator = equilibrium_indicator | TEMPERATURE_STABLE

        if self.humidity_stable:
            equilibrium_indicator = equilibrium_indicator | HUMIDITY_STABLE

        if self.magnitude_stable:
            equilibrium_indicator = equilibrium_indicator | MAGNITUDE_STABLE

        if self.phase_stable:
            equilibrium_indicator = equilibrium_indicator | PHASE_STABLE

        return equilibrium_indicator

    def calculate_temperature_stability(self, target_temp, readback_temp):
        return (target_temp-self.max_delta_temp <= float(readback_temp)) and \
               (float(readback_temp) <= target_temp+self.max_delta_temp)

    def calculate_humidity_stability(self, target_hum, readback_hum):
        return (target_hum-self.max_delta_hum <= float(readback_hum)) and \
               (float(readback_hum) <= target_hum+self.max_delta_hum)

def plot_output(output_basename, measurements_appendices, show_blocking_plot, config_data, ext_sens_data, title = ''):
    
    # set time unit for PostPlot of measurement results to imported value from json file
    # test_stand_parameter.json, Michael
    time_unit = str(config_data['time_unit'])                

    list_of_frames = []
    # storepath is set to working directory with subfolder Plots, Michael
    storepath = os.path.join(os.getcwd(),'Plots')
    # create objet for PostPlot class in PostPlot module, Michael
    post_plot = PostPlot.PostPlot()
    for index, m in enumerate(measurements_appendices):
        measurement_name = output_basename+'_'+str(m)
        list_of_frames.append(pd.read_csv(measurement_name+'.csv'))

      
        # plot frame data
        post_plot.plot_frame_data(list_of_frames[-1], title, time_unit,)  
        
        # store generated postplot to the in storepath defined destination, Michael
        post_plot.save_fig(storepath, measurement_name+'.pdf')

    combined_data_frame = pd.concat(list_of_frames, ignore_index=True, sort=False)
    
    # create postplot of all measurements, Michael
    # add the imported time unit from ext_sens_data.json to parameterlist   
    post_plot.plot_frame_data(combined_data_frame, title, time_unit)
    
    # save postplot of all measurents to the in storpath defined destination and name it FullTransistion
    post_plot.save_fig(storepath, 'FullTransistion.pdf')
    
    if show_blocking_plot:
        plt.ioff()
    plot = MeasurementPlot.MeasurementPlot(title)
    plot.draw(combined_data_frame, output_basename + '_graph.pdf')
    

def run_temperature_sweep_from_file(temperature_sweep_file, meas):
    with open(temperature_sweep_file) as file:
        try:
            line = file.readline().rstrip()
            list_of_values = line.split()
            start_temp = float(list_of_values[0])
            stop_temp = float(list_of_values[1])
            temperature_step = float(list_of_values[2])
            target_hum = float(list_of_values[3])
            soaking_time = int(list_of_values[4])
            n_stable_reads = int(list_of_values[5])

            return meas.perform_sweep(start_temp, stop_temp, temperature_step, target_hum, soaking_time,
                                      n_stable_reads, 'temperature'), target_hum

        except KeyboardInterrupt:
            pass

if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument("-c", "--chamber",
                        help="IP address of climate chamber", metavar="ADDR",
                        required=True)
    parser.add_argument("-v", "--vna",
                        help="IP address of VNA", metavar="ADDR",
                        required=True)
    parser.add_argument('-f', '--file',
                        help='File containing custom list of measurements',
                        default='')
    parser.add_argument('-t', '--temperaturesweepfile',
                        help='File containing temperature sweep parameters',
                        default='')
    parser.add_argument('-o', '--output',
                        help='Name of output file')
    parser.add_argument("-s", "--standby",
                        action="store_true", default=False,
                        help="Change Chamber mode to standby at the end")
    parser.add_argument("-p", "--plot",
                        action="store_true", default=False,
                        help="Show blocking plot window at the end")

    # parser.add_argument("-q", "--quiet",
    #                     action="store_false", dest="verbose", default=True,
    #                     help="don't print status messages to stdout")
    args = parser.parse_args()

    if args.file == '' and args.temperaturesweepfile == '':
        print('Argument error: Either \'file\' or \'temperaturesweepfile\' must be specified.')
        sys.exit(-1)

    if not args.output:
        output_basename = time.strftime("%Y_%m_%d-%H_%M_%S") + "_results"
    else:
        output_basename = args.output

    print(args.chamber, args.vna, args.file, output_basename, args.standby)

    # reading json file for target accuracy
    with open('test_stand_parameter.json', 'r') as f:
        config_data = json.load(f)

    mes = Measurements(args.chamber, args.vna, output_basename, args.standby, config_data)
    try:
        if args.file:
            n_measurements = mes.perform_measurements(args.file)
            plot_output(output_basename, range(n_measurements), args.plot, output_basename)
        if args.temperaturesweepfile:
            temperatures, humidity = run_temperature_sweep_from_file(args.temperaturesweepfile, mes)
            #run analysis here
            temp_extensions = []
            for t in temperatures:
                temp_extensions.append(str(t)+'deg_'+str(humidity)+'rh')
            analysis.plot_sweep(temperatures, [humidity]*len(temperatures), output_basename, 'temperature')
            plot_output(output_basename, temp_extensions, args.plot, output_basename + ': Temperature sweep ' +
                        str(temperatures[0]) + ' degC to ' + str(temperatures[-1]) + ' degC')
            print(str(temp_extensions))
    finally:
        mes.chamber.close()
        mes.ext_sensors.close()

    # mes.plot_output(output)

